// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_CURVED_SURFACE_GRID_GRID_HH
#define DUNE_CURVED_SURFACE_GRID_GRID_HH

#include <dune/common/deprecated.hh>
#include <dune/common/shared_ptr.hh>

#include <dune/grid/common/grid.hh>

#include <dune/curvedsurfacegrid/concepts.hh>
#include <dune/curvedsurfacegrid/gridfamily.hh>
#include <dune/curvedsurfacegrid/backuprestore.hh>
#include <dune/curvedsurfacegrid/datahandle.hh>
#include <dune/curvedsurfacegrid/gridfunctions/analyticgridfunction.hh>
#include <dune/curvedsurfacegrid/gridfunctions/gridfunction.hh>
#include <dune/functions/common/functionconcepts.hh>
#include <dune/grid/geometrygrid/identity.hh>
#include <dune/grid/geometrygrid/persistentcontainer.hh>
#include <dune/grid/geometrygrid/grid.hh>

namespace Dune
{
  namespace Impl
  {
    template< class GridFunction, int order >
    struct CurvedSurfaceGridBase
    {
      using HG = GridOf_t<GridFunction>;

      using type = GridDefaultImplementation<
        HG::dimension, Curved::DimRange<GridFunction>::value, typename HG::ctype,
        Curved::GridFamily<GridFunction,order>
      >;
    };
  }

  template< class GridFunction, int order >
  using CurvedSurfaceGridBase = typename Impl::CurvedSurfaceGridBase<GridFunction,order>::type;


  // CurvedSurfaceGrid
  // -----------------

  /** \class CurvedSurfaceGrid
   *  \brief grid wrapper replacing the geometries
   *  \ingroup CurvedGeo
   *
   *  CurvedSurfaceGrid wraps another DUNE grid and replaces its geometry by an implementation
   *  of a CurvedGeometry.
   *
   *  \tparam GF     GridViewFunction defined on a (flat) hostgrid
   *  \tparam ORDER  Polynomial order of local lagrange basis functions to use when approximating
   *                 the curved geometry. [optional]
   *
   *  \nosubgrouping
   */
  template< class GF, int ORDER = -1 >
  class CurvedSurfaceGrid;

  //! Generator for CurvedSurfaceGrid from a grid-functions
  template< class HostGrid, class GF, int ORDER = -1, 
    std::enable_if_t<Dune::Concept::isGridFunction<GF, HostGrid>(), int> = 0>
  auto curedSurfaceGrid (HostGrid& hostGrid, GF&& gridFunction) 
  {
    static_assert(std::is_same<HostGrid, GridOf_t<std::decay_t<GF>>>::value, "GridFunction must be defined on the HostGrid");
    return CurvedSurfaceGrid<std::decay_t<GF>,ORDER>{hostGrid, std::forward<GF>(gridFunction)};
  }

  //! Generator for CurvedSurfaceGrid from a callable
  template< class HostGrid, class F, int ORDER = -1, 
    std::enable_if_t<not Dune::Concept::isGridFunction<F, HostGrid>(), int> = 0>
  auto curedSurfaceGrid (HostGrid& hostGrid, F&& callable) 
  {
    using GlobalCoordinate = typename GridEntitySet<HostGrid,0>::GlobalCoordinate;
    static_assert(Functions::Concept::isCallable<F, GlobalCoordinate>(), "Function must be callable");
    auto gridFct = analyticGridFunction<HostGrid>(std::forward<F>(callable));
    return CurvedSurfaceGrid<decltype(gridFct),ORDER>{hostGrid, std::move(gridFct)};
  }


  template< class GF, int ORDER >
  class CurvedSurfaceGrid
      : public CurvedSurfaceGridBase<GF,ORDER>
      , public Curved::BackupRestoreFacilities<CurvedSurfaceGrid<GF,ORDER> >
  {
    using Self = CurvedSurfaceGrid;
    using Super = CurvedSurfaceGridBase<GF,ORDER>;

    // friend declarations
    template< class, class > friend class Curved::IdSet;
    template< class, class > friend class Curved::IndexSet;

  public:
    using GridFunction = GF;
    using GridFamily = Curved::GridFamily<GF,ORDER>;

    /** \name Traits
     *  \{ */

    //! type of the grid traits
    using Traits = typename GridFamily::Traits;

    using HostGrid = typename Traits::HostGrid;

    //! traits structure containing types for a codimension
    /**
     *  \tparam codim  codimension
     *
     *  \nosubgrouping
     */
    template< int codim >
    struct Codim;

    /** \} */

    /** \name Iterator Types
     *  \{ */

    //! iterator over the grid hierarchy
    using HierarchicIterator = typename Traits::HierarchicIterator;

    //! iterator over intersections with other entities on the leaf level
    using LeafIntersectionIterator = typename Traits::LeafIntersectionIterator;

    //! iterator over intersections with other entities on the same level
    using LevelIntersectionIterator = typename Traits::LevelIntersectionIterator;

    /** \} */


    /** \name Grid View Types
     *  \{ */

    //! type of view for leaf grid
    using LeafGridView = typename GridFamily::Traits::LeafGridView;

    //! type of view for level grid
    using LevelGridView = typename GridFamily::Traits::LevelGridView;

    /** \} */


    /** \name Index and Id Set Types
     *  \{ */

    /** \brief type of leaf index set
     *
     *  The index set assigns consecutive indices to the entities of the
     *  leaf grid. The indices are of integral type and can be used to access
     *  arrays.
     *
     *  The leaf index set is a model of Dune::IndexSet.
     */
    using LeafIndexSet = typename Traits::LeafIndexSet;

    /** \brief type of level index set
     *
     *  The index set assigns consecutive indices to the entities of a grid
     *  level. The indices are of integral type and can be used to access
     *  arrays.
     *
     *  The level index set is a model of Dune::IndexSet.
     */
    using LevelIndexSet = typename Traits::LevelIndexSet;

    /** \brief type of global id set
     *
     *  The id set assigns a unique identifier to each entity within the
     *  grid. This identifier is unique over all processes sharing this grid.
     *
     *  \note Id's are neither consecutive nor necessarily of an integral
     *        type.
     *
     *  The global id set is a model of Dune::IdSet.
     */
    using GlobalIdSet = typename Traits::GlobalIdSet;

    /** \brief type of local id set
     *
     *  The id set assigns a unique identifier to each entity within the
     *  grid. This identifier needs only to be unique over this process.
     *
     *  Though the local id set may be identical to the global id set, it is
     *  often implemented more efficiently.
     *
     *  \note Ids are neither consecutive nor necessarily of an integral
     *        type.
     *  \note Local ids need not be compatible with global ids. Also, no
     *        mapping from local ids to global ones needs to exist.
     *
     *  The global id set is a model of Dune::IdSet.
     */
    using LocalIdSet = typename Traits::LocalIdSet;

    /** \} */

    /** \name Miscellaneous Types
     * \{ */

    //! type of vector coordinates (e.g., double)
    using ctype = typename Traits::ctype;

    //! communicator with all other processes having some part of the grid
    using CollectiveCommunication = typename Traits::CollectiveCommunication;

    /** \} */

    /** \name Construction and Destruction
     *  \{ */

    //! constructor for host-grid given by reference
    /**
     *  The references to host grid is stored in the grid. It must be valid until this grid
     *  wrapper is destroyed.
     *
     *  \param[in]  hostGrid      reference to the grid to wrap
     *  \param[in]  gridFunction  mapping from global coordinates in the host geometry
     *                            to global coordinates in the curved geometry
     * 
     *  If the gridFunction is passed by (non-const) reference it is stored as a non-destroying shared_ptr.
     *  Otherwise it is copied or moved into a new object (stored as shared_ptr as well).
     */
    template <class GF_,
      std::enable_if_t<std::is_same<GF,std::decay_t<GF_>>::value, int> = 0>
    CurvedSurfaceGrid (HostGrid& hostGrid, GF_&& gridFunction, std::integral_constant<int,ORDER> = {})
      : hostGrid_(hostGrid)
      , gridFunction_(wrap_or_move(std::forward<GF_>(gridFunction)))
      , levelIndexSets_(hostGrid_.maxLevel()+1, nullptr)
    {}

    template <class F_,
      std::enable_if_t<std::is_same<GF,AnalyticGridFunction<HostGrid,std::decay_t<F_>>>::value, int> = 0>
    CurvedSurfaceGrid (HostGrid& hostGrid, F_&& callable, std::integral_constant<int,ORDER> = {})
      : CurvedSurfaceGrid(hostGrid, analyticGridFunction<HostGrid>(std::forward<F_>(callable)))
    {}

    //! destructor
    ~CurvedSurfaceGrid ()
    {
      for (unsigned int i = 0; i < levelIndexSets_.size(); ++i)
      {
        if (levelIndexSets_[i])
          delete (levelIndexSets_[i]);
      }
    }

    /** \} */

    /** \name Size Methods
     *  \{ */

    //! obtain maximal grid level
    /**
     *  Grid levels are numbered 0, ..., L, where L is the value returned by
     *  this method.
     *
     *  \returns maximal grid level
     */
    int maxLevel () const
    {
      return hostGrid().maxLevel();
    }

    //! obtain number of entites on a level
    /**
     *  \param[in]  level  level to consider
     *  \param[in]  codim  codimension to consider
     *
     *  \returns number of entities of codimension `codim` on grid level `level`.
     */
    int size (int level, int codim) const
    {
      return levelGridView(level).size(codim);
    }

    //! obtain number of leaf entities
    /**
     *  \param[in]  codim  codimension to consider
     *
     *  \returns number of leaf entities of codimension \em codim
     */
    int size (int codim) const
    {
      return leafGridView().size(codim);
    }

    //! obtain number of entites on a level
    /**
     *  \param[in]  level  level to consider
     *  \param[in]  type   geometry type to consider
     *
     *  \returns number of entities with a geometry of type `type` on grid level `level`.
     */
    int size (int level, GeometryType type) const
    {
      return levelGridView(level).size(type);
    }

    //! obtain number of leaf entities
    /**
     *  \returns number of leaf entities with a geometry of type `type`
     */
    int size (GeometryType type) const
    {
      return leafGridView().size(type);
    }

    //! returns the number of boundary segments within the macro grid
    /**
     *  \returns number of boundary segments within the macro grid
     */
    std::size_t numBoundarySegments () const
    {
      return hostGrid().numBoundarySegments();
    }
    /** \} */

    const GlobalIdSet& globalIdSet () const
    {
      if (!globalIdSet_)
        globalIdSet_ = GlobalIdSet(hostGrid().globalIdSet());
      assert(globalIdSet_);
      return globalIdSet_;
    }

    const LocalIdSet& localIdSet () const
    {
      if (!localIdSet_)
        localIdSet_ = LocalIdSet(hostGrid().localIdSet());
      assert(localIdSet_);
      return localIdSet_;
    }

    const LevelIndexSet& levelIndexSet (int level) const
    {
      assert(levelIndexSets_.size() == (size_t)(maxLevel()+1));

      if ((level < 0) || (level > maxLevel()))
        DUNE_THROW(GridError, "LevelIndexSet for nonexisting level " << level << " requested.");

      LevelIndexSet*& levelIndexSet = levelIndexSets_[level];
      if (!levelIndexSet)
        levelIndexSet = new LevelIndexSet(hostGrid().levelIndexSet(level));
      assert(levelIndexSet);
      return *levelIndexSet;
    }

    const LeafIndexSet& leafIndexSet () const
    {
      if (!leafIndexSet_)
        leafIndexSet_.reset(hostGrid().leafIndexSet());
      assert(leafIndexSet_);
      return leafIndexSet_;
    }

    void globalRefine (int refCount)
    {
      hostGrid().globalRefine(refCount);
      update();
    }

    bool mark (int refCount, const typename Codim<0>::Entity& entity)
    {
      return hostGrid().mark(refCount, getHostEntity<0>(entity));
    }

    int getMark (const typename Codim<0>::Entity& entity) const
    {
      return hostGrid().getMark(getHostEntity<0>(entity));
    }

    bool preAdapt ()
    {
      return hostGrid().preAdapt();
    }

    bool adapt ()
    {
      bool ret = hostGrid().adapt();
      update();
      return ret;
    }

    void postAdapt ()
    {
      hostGrid().postAdapt();
    }

    /** \name Parallel Data Distribution and Communication Methods
     *  \{ */

    //! obtain CollectiveCommunication object
    /**
     *  The CollectiveCommunication object should be used to globally
     *  communicate information between all processes sharing this grid.
     *
     *  \note The CollectiveCommunication object returned is identical to the
     *        one returned by the host grid.
     */
    const CollectiveCommunication& comm () const
    {
      return hostGrid().comm();
    }

    // data handle interface different between geo and interface

    //! rebalance the load each process has to handle
    /**
     *  A parallel grid is redistributed such that each process has about
     *  the same load (e.g., the same number of leaf entites).
     *
     *  \note DUNE does not specify, how the load is measured.
     *
     *  \returns \b true, if the grid has changed.
     */
    bool loadBalance ()
    {
      const bool gridChanged = hostGrid().loadBalance();
      if (gridChanged)
        update();
      return gridChanged;
    }

    //! rebalance the load each process has to handle
    /**
     *  A parallel grid is redistributed such that each process has about
     *  the same load (e.g., the same number of leaf entites).
     *
     *  The data handle is used to communicate the data associated with
     *  entities that move from one process to another.
     *
     *  \note DUNE does not specify, how the load is measured.
     *
     *  \param  datahandle  communication data handle (user defined)
     *
     *  \returns \b true, if the grid has changed.
     */

    template< class DataHandle, class Data >
    bool loadBalance (CommDataHandleIF<DataHandle, Data>& datahandle)
    {
      using DataHandleIF = CommDataHandleIF<DataHandle, Data>;
      using WrappedDataHandle = Curved::CommDataHandle<Self, DataHandleIF>;

      WrappedDataHandle wrappedDataHandle(*this, datahandle);
      const bool gridChanged = hostGrid().loadBalance(wrappedDataHandle);
      if (gridChanged)
        update();
      return gridChanged;
    }

    //! obtain Entity from EntitySeed
    /**
     *  EntitySeed survives to a grid modification which only changes the grid coordinates.
     *  Therefore it is consistent to use an EntitySeed to rebuild an Entity after this kind
     *  of grid modification.
     */
    template< class EntitySeed >
    typename Traits::template Codim<EntitySeed::codimension>::Entity entity (const EntitySeed& seed) const
    {
      using EntityImpl = typename Traits::template Codim<EntitySeed::codimension>::EntityImpl;
      return EntityImpl(*this, seed);
    }

    /** \} */

    /** \name Grid Views
     *  \{ */

    //! View for a grid level
    LevelGridView levelGridView (int level) const
    {
      using ViewImp = typename LevelGridView::GridViewImp;
      return LevelGridView(ViewImp(*this, hostGrid().levelGridView(level)));
    }

    //! View for the leaf grid
    LeafGridView leafGridView () const
    {
      using ViewImp = typename LeafGridView::GridViewImp;
      return LeafGridView(ViewImp(*this, hostGrid().leafGridView()));
    }

    /** \} */

    /** \name Miscellaneous Methods
     *  \{ */

    //! obtain constant reference to the host grid
    const HostGrid& hostGrid () const
    {
      return hostGrid_;
    }

    //! obtain mutable reference to the host grid
    HostGrid& hostGrid ()
    {
      return hostGrid_;
    }

    //! update grid caches
    /**
     *  This method has to be called whenever the underlying host grid changes.
     *
     *  \note If you adapt the host grid through this geometry grid's
     *        adaptation or load balancing methods, update is automatically
     *        called.
     */
    void update ()
    {
      // adapt the coordinate function
      // GeoGrid::AdaptCoordFunction<typename CoordFunction::Interface>::adapt(coordFunction());

      const int newNumLevels = maxLevel()+1;
      const int oldNumLevels = levelIndexSets_.size();

      for (int i = newNumLevels; i < oldNumLevels; ++i)
      {
        if (levelIndexSets_[i])
          delete levelIndexSets_[i];
      }
      levelIndexSets_.resize(newNumLevels, nullptr);
    }

    //! obtain constant reference to the coordinate function
    GridFunction const& gridFunction () const 
    { 
      return *gridFunction_; 
    }

    GridFunction& gridFunction () 
    { 
      return *gridFunction_; 
    }

    bool useGeometryCaching () const
    {
      return useGeometryCaching_;
    }

    void setGeometryCaching (bool cache)
    {
      useGeometryCaching_ = cache;
    }

    /** \} */

  protected:
    template< int codim >
    static const typename HostGrid::template Codim<codim>::Entity&
    getHostEntity(const typename Codim<codim>::Entity& entity)
    {
      return entity.impl().hostEntity();
    }

  private:
    HostGrid& hostGrid_;
    std::shared_ptr<GridFunction> gridFunction_;

    bool useGeometryCaching_ = false;

    mutable std::vector<LevelIndexSet*> levelIndexSets_;
    mutable LeafIndexSet leafIndexSet_;
    mutable GlobalIdSet globalIdSet_;
    mutable LocalIdSet localIdSet_;

  private:
    template <class Indices>
    struct GeometryCache;

    template <int... codims>
    struct GeometryCache<std::integer_sequence<int,codims...>>
    {
      using type = std::tuple<
        std::map<
          typename HostGrid::Traits::LocalIdSet::IdType,
          typename Traits::template Codim<codims>::GeometryImpl
        >...
      >;
    };

    mutable typename GeometryCache<std::make_integer_sequence<int,Traits::dimension+1>>::type geometryCache_;
  };

  //! Deduction guide for CurvedSurfaceGrid from a grid-functions or callables
  template< class HostGrid, class GF, int ORDER = -1 >
  CurvedSurfaceGrid (HostGrid& hostGrid, GF&& gridFunction, std::integral_constant<int,ORDER> = {}) 
    -> CurvedSurfaceGrid<GridFunctionOf_t<HostGrid,std::decay_t<GF>>,ORDER>;


  // CurvedSurfaceGrid::Codim
  // ------------------------

  template< class GridFunction, int ORDER >
  template< int codim >
  struct CurvedSurfaceGrid<GridFunction,ORDER>::Codim
    : public Super::template Codim< codim >
  {
    /** \name Entity types
     *  \{ */

    /** \brief type of entity
     *
     *  The entity is a model of Dune::Entity.
     */
    using Entity = typename Traits::template Codim< codim >::Entity;

    /** \} */

    /** \name Geometry Types
     *  \{ */

    /** \brief type of world geometry
     *
     *  Models the geomtry mapping of the entity, i.e., the mapping from the
     *  reference element into world coordinates.
     *
     *  The geometry is a model of Dune::Geometry, implemented through the
     *  generic geometries provided by dune-grid.
     */
    using Geometry = typename Traits::template Codim< codim >::Geometry;

    /** \brief type of local geometry
     *
     *  Models the geomtry mapping into the reference element of dimension
     *  \em dimension.
     *
     *  The local geometry is a model of Dune::Geometry, implemented through
     *  the generic geometries provided by dune-grid.
     */
    using LocalGeometry = typename Traits::template Codim< codim >::LocalGeometry;

    /** \} */

    /** \name Iterator Types
     *  \{ */

    template< PartitionIteratorType pitype >
    struct Partition
    {
      using LeafIterator
        = typename Traits::template Codim<codim>::template Partition< pitype >::LeafIterator;
      using LevelIterator
        = typename Traits::template Codim<codim>::template Partition< pitype >::LevelIterator;
    };

    /** \brief type of leaf iterator
     *
     *  This iterator enumerates the entites of codimension \em codim of a
     *  grid level.
     *
     *  The level iterator is a model of Dune::LevelIterator.
     */
    using LeafIterator = typename Partition< All_Partition >::LeafIterator;

    /** \brief type of level iterator
     *
     *  This iterator enumerates the entites of codimension \em codim of the
     *  leaf grid.
     *
     *  The leaf iterator is a model of Dune::LeafIterator.
     */
    using LevelIterator = typename Partition< All_Partition >::LevelIterator;

    /** \} */
  };

} // namespace Dune

#endif // DUNE_CURVED_SURFACE_GRID_GRID_HH
