#include <dune/curvedsurfacegrid/surfacedistance/vtureader.hh>
#ifdef DCSG_BOOST_IOSTREAMS
  #include <boost/iostreams/device/array.hpp>
  #include <boost/iostreams/copy.hpp>
  #include <boost/iostreams/filtering_streambuf.hpp>
  #include <boost/iostreams/filtering_stream.hpp>
  #include <boost/iostreams/filter/zlib.hpp>
  #include <boost/iostreams/filter/bzip2.hpp> //bzip2 is slower but compresses better, usually.
#endif

using namespace Dune::Curved;

//--------------------< decompress_buffer >
//Use this routine if the size of the uncompressed data 'osize' is known. Decompresses buffer 'ibuf'
//of size 'isize' into an already allocated buffer 'obuf' of size 'osize'.
#ifdef DCSG_BOOST_IOSTREAMS
void decompress_buffer(const char *ibuf, const unsigned isize, char *obuf, const unsigned osize,
                                                                      const int compression_type){
  boost::iostreams::array_source source(ibuf, isize);
  boost::iostreams::array_sink sink(obuf, osize);
  boost::iostreams::filtering_streambuf<boost::iostreams::input> fin;
  if(compression_type == 1){
    fin.push(boost::iostreams::zlib_decompressor());
  }else if(compression_type == 2){
    fin.push(boost::iostreams::bzip2_decompressor());
  }
  fin.push(source);
  boost::iostreams::copy(fin, sink);
}
#endif

//--------------------< read_compressed_buffer >
//Use this routine if the size of the uncompressed data 'bsize' is known. Reads 'fsize' bytes from
//'in' and decompresses them into the already allocated buffer 'buf' of size 'bsize'.
#ifdef DCSG_BOOST_IOSTREAMS
void read_compressed_buffer(std::ifstream &in, char *buf, const unsigned bsize,
                               const unsigned fsize, const int compression_type){
  char *ibuf = new char[fsize];
  in.read(ibuf, fsize);
  decompress_buffer(ibuf, fsize, buf, bsize, compression_type);
  delete [] ibuf;
}
#endif

//--------------------< VtuValueReader >
struct VtuValueReader{
  int64_t total_size, block_cnt, block_size, last_block_size, block_index, byte_position;
  std::ifstream *in;
  char *cbuf;
  int header_size;

  VtuValueReader(std::ifstream &in, int header_size=8):block_size(0), block_index(-1),
                 byte_position(0), in(&in), cbuf(NULL), header_size(header_size){/*$ACL_exclude*/}

  virtual ~VtuValueReader(){//$ACL_exclude
    if(cbuf != NULL) delete [] cbuf;
  }

  virtual void initialize(int64_t block_size=32768){//$ACL_exclude
    if(this->block_size != block_size){
      if(cbuf != NULL) delete [] cbuf;
      cbuf = new char[block_size];
      this->block_size = block_size;
    }
    if(header_size == 4){
      int32_t tmp;
      in->read(reinterpret_cast<char*>(&tmp), 4); //total size
      total_size = tmp;
    }else if(header_size == 8){
      in->read(reinterpret_cast<char*>(&total_size), 8); //total size
    }else{
      total_size = 0;
      error("Illeagal header_type in VtuValueReader!");
    }
    block_cnt = total_size / block_size;
    last_block_size = total_size % block_size;
    if(last_block_size == 0) last_block_size = block_size;
    else ++block_cnt;
    block_index = -1;
    get_next_block();
  }

  virtual void get_next_block(){//$ACL_exclude
    if(++block_index == block_cnt - 1) in->read(cbuf, last_block_size);
    else in->read(cbuf, block_size);
    byte_position = 0;
  }

  virtual double read_double_value(){//$ACL_exclude
    if(byte_position == block_size) get_next_block();
    int64_t index = byte_position;
    byte_position += 8;
    return *reinterpret_cast<double*>(&cbuf[index]);
  }

  virtual float read_float_value(){//$ACL_exclude
    if(byte_position == block_size) get_next_block();
    int64_t index = byte_position;
    byte_position += 4;
    return *reinterpret_cast<float*>(&cbuf[index]);
  }

  virtual int64_t read_int64_value(){//$ACL_exclude
    if(byte_position == block_size) get_next_block();
    int64_t index = byte_position;
    byte_position += 8;
    return *reinterpret_cast<int64_t*>(&cbuf[index]);
  }

  virtual int32_t read_int32_value(){//$ACL_exclude
    if(byte_position == block_size) get_next_block();
    int64_t index = byte_position;
    byte_position += 4;
    return *reinterpret_cast<int32_t*>(&cbuf[index]);
  }

  virtual uchar read_uchar_value(){//$ACL_exclude
    if(byte_position == block_size) get_next_block();
    int64_t index = byte_position;
    ++byte_position;
    return *reinterpret_cast<uchar*>(&cbuf[index]);
  }
};

//--------------------< VtuValueDecompressor >
struct VtuValueDecompressor : public VtuValueReader{
  std::vector<int64_t> block_sizes_compressed;

  VtuValueDecompressor(std::ifstream &in, int header_size=8):VtuValueReader(in, header_size){
                                                                             /*$ACL_exclude*/}

  virtual void initialize(int64_t block_size=32768){ //$ACL_exclude
    //note: block-size-argument is not used here.
    if(header_size == 4){
      uint32_t tmp;
      in->read(reinterpret_cast<char*>(&tmp), 4); //number of blocks (including last one)
      block_cnt = tmp;
      in->read(reinterpret_cast<char*>(&tmp), 4); //uncompressed block-size
      block_size = tmp;
      in->read(reinterpret_cast<char*>(&tmp), 4); //uncompressed last block-size
      last_block_size = tmp;
    }else if(header_size == 8){
      in->read(reinterpret_cast<char*>(&block_cnt), 8); //number of blocks (including last one)
      in->read(reinterpret_cast<char*>(&block_size), 8); //uncompressed block-size
      in->read(reinterpret_cast<char*>(&last_block_size), 8); //uncompressed last block-size
    }else{
      block_cnt = 0;
      block_size = 0;
      last_block_size = 0;
      error("Illegal header_type in VtuValueDecompressor!");
    }
    if(last_block_size == 0) last_block_size = block_size;
    total_size = (block_cnt - 1) * block_size + last_block_size;
    block_sizes_compressed.resize(block_cnt);
    for(unsigned i = 0; i < block_cnt; ++i){
      if(header_size == 4){
        uint32_t tmp;
        in->read(reinterpret_cast<char*>(&tmp), 4); //compressed block-sizes
        block_sizes_compressed[i] = tmp;
      }else if(header_size == 8){
        in->read(reinterpret_cast<char*>(&block_sizes_compressed[i]), 8); //compressed block-sizes
      }
    }
    if(this->block_size != block_size){
      if(cbuf != NULL) delete [] cbuf;
      cbuf = new char[block_size];
      this->block_size = block_size;
    }
    block_index = -1;
    get_next_block();
  }

  virtual void get_next_block(){ //$ACL_exclude
    #ifdef DCSG_BOOST_IOSTREAMS
      if(++block_index == block_cnt - 1){
        read_compressed_buffer(*in, cbuf, last_block_size, block_sizes_compressed[block_cnt - 1], 1);
      }else{
        read_compressed_buffer(*in, cbuf, block_size, block_sizes_compressed[block_index], 1);
      }
      byte_position = 0;
    #else
      error("dune-curvedsurfacegrid was not compiled with compression-support!");
    #endif
  }
};

//--------------------< read_vtu_file >
//triangular meshes in 2d- and 3d-worlds and tetrahedral meshes in 3d worlds supported at the moment
void Dune::Curved::read_vtu_file(std::string file, Mesh &m, std::vector<DataVector*> &vecs,
                                                     std::vector<DataVector*> &element_vecs){
  std::string buf = "";
  std::string connectivity_buf;
  int v[4];
  int header_size;
  unsigned vertex_count = 0, component_count;
  size_t pos1, pos2;
  uint64_t position_of_cells = 0, position_of_connectivity = 0;
  uint64_t position_of_element_vectors = 0, position_of_vectors = 0;
  uint64_t element_type;
  Vertex vc;
  int64_t offset_points = -1;
  int64_t offset_cell_types = -1;
  int64_t offset_cell_connectivity = -1;
  std::vector<int64_t> offsets_point_data;
  std::vector<int64_t> offsets_cell_data;
  std::string datatype_points;
  std::string datatype_cell_types;
  std::string datatype_cell_connectivity;
  std::vector<std::string> datatypes_point_data;
  std::vector<std::string> datatypes_cell_data;
  std::vector<unsigned> component_counts_point_data;
  std::vector<unsigned> component_counts_cell_data;
  bool has_compression = false;
  bool has_appended_data = false;

  //open input file
  std::ifstream in(file.c_str(), std::ios_base::in | std::ios_base::binary);
  if(!in) error("Could not read file \"" + file + "\".");

  //get compression-type
  do{
    getline(in, buf);
    if(in.eof()) error("File does not have a valid vtu-header.");
    buf = ltrim(buf);
  }while(buf.size() < 8 || buf.substr(0, 8) != "<VTKFile");
  if(buf.find("compressor=\"vtkZLibDataCompressor\"") != npos){
    has_compression = true;
  }
  if(buf.find("header_type=\"UInt64\"") != npos){
    header_size = 8;
  }else{
    header_size = 4;
  }

  //get vertices
  do{
    getline(in, buf);
    if(in.eof()) error("Did not find vertices.");
    buf = ltrim(buf);
    if(buf.size() >= 7 && buf.substr(0, 7) == "<Cells>") position_of_cells = in.tellg();
    if(buf.size() >= 10 && buf.substr(0, 10) == "<PointData") position_of_vectors = in.tellg();
    if(buf.size() >= 9 && buf.substr(0, 9) == "<CellData") position_of_element_vectors = in.tellg();
  }while(buf.size() < 8 || buf.substr(0, 8) != "<Points>");
  do{
    getline(in, buf);
    if(in.eof()) error("Did not find vertices.");
    buf = ltrim(buf);
  }while(buf.size() < 10 || buf.substr(0, 10) != "<DataArray");
  if(buf.find("\"binary\"") != npos){
    error("dune-curvedsurfacegrid cannot handle binary-mode .vtu-files. Convert the "
                                 "file to appended-mode or ascii-mode and try again.");
  }else if(buf.find("\"appended\"") != npos){
    has_appended_data = true;
    pos1 = buf.find("offset=\"");
    if(pos1 == npos) error("Could not find offset for appended data of \"Points\".");
    else pos1 += 8;
    pos2 = buf.find("\"", pos1);
    if(pos2 == npos) error("Could not find offset for appended data of \"Points\".");
    offset_points = ival64(buf.substr(pos1, pos2 - pos1));
    pos1 = buf.find("type=\"");
    if(pos1 == npos) error("Could not find datatype for appended data of \"Points\".");
    else pos1 += 6;
    pos2 = buf.find("\"", pos1);
    if(pos2 == npos) error("Could not find datatype for appended data of \"Points\".");
    datatype_points = buf.substr(pos1, pos2 - pos1);
  }else{
    unsigned i = 0;
    do{
      in >> buf;
      buf = ltrim(buf);
      if(buf.size() > 0 && buf[0] == '<') break;
      vc[i] = dval(buf);
      i = (i + 1) % 3; //always read three coordinates even if dimension of world is 2!
      if(i == 0) m.add_vertex(vc);
    }while(true);
  }

  //get elements
  vertex_count = m.vertices.size();
  if(position_of_cells != 0){
    in.seekg(position_of_cells);
  }else{
    do{
      getline(in, buf);
      if(in.eof()) error("Did not find elements.");
      buf = ltrim(buf);
      if(buf.size() >= 10 && buf.substr(0, 10) == "<PointData") position_of_vectors = in.tellg();
      if(buf.size() >= 9 && buf.substr(0, 9) == "<CellData"){
        position_of_element_vectors = in.tellg();
      }
    }while(buf.size() < 7 || buf.substr(0,7) != "<Cells>");
  }

  //search element-types
  do{
    getline(in, buf);
    buf = ltrim(buf);
    if(in.eof()) error("Did not find elements.");
    pos1 = buf.find("\"connectivity\"");
    if(pos1 != npos){
      connectivity_buf = buf;
      position_of_connectivity = in.tellg();
    }
    pos1 = buf.find("\"types\"");
  }while(pos1 == npos);
  if(buf.find("\"binary\"") != npos){
    error("dune-curvedsurfacegrid cannot handle binary-mode .vtu-files. Convert the "
                                 "file to appended-mode or ascii-mode and try again.");
  }else if(buf.find("\"appended\"") != npos){
    has_appended_data = true;
    pos1 = buf.find("offset=\"");
    if(pos1 == npos) error("Could not find offset for appended data of \"types\".");
    else pos1 += 8;
    pos2 = buf.find("\"", pos1);
    if(pos2 == npos) error("Could not find offset for appended data of \"types\".");
    offset_cell_types = ival64(buf.substr(pos1, pos2 - pos1));
    pos1 = buf.find("type=\"");
    if(pos1 == npos) error("Could not find datatype for appended data of \"types\".");
    else pos1 += 6;
    pos2 = buf.find("\"", pos1);
    if(pos2 == npos) error("Could not find datatype for appended data of \"types\".");
    datatype_cell_types = buf.substr(pos1, pos2 - pos1);
  }else{
    in >> buf;
    if(buf == "</DataArray>"){
      m.dim_of_elements = 2;
    }else{
      element_type = ival64(buf);
      if(element_type == 3) m.dim_of_elements = 1;
      else if(element_type == 5) m.dim_of_elements = 2;
      else if(element_type == 10) m.dim_of_elements = 3;
      else error("Unknown element type " + ft(int(element_type)) + ".");
      unsigned tmp_type;
      while(true){
        in >> buf;
        if(buf.size() > 0 && buf[0] == '<') break;
        tmp_type = ival(buf);
        if(tmp_type != element_type){
          error("File contains multiple element types! At least "
                  + ft((unsigned)element_type) + " and " + ft(tmp_type) + ".");
        }
      }
    }
  }

  //search connectivity
  if(position_of_connectivity != 0){
    in.seekg(position_of_connectivity);
  }else{
    do{
      getline(in, buf);
      buf = ltrim(buf);
      if(in.eof()) error("Did not find elements.");
      pos1 = buf.find("\"connectivity\"");
    }while(pos1 == npos);
    connectivity_buf = buf;
  }
  if(connectivity_buf.find("\"binary\"") != npos){
    error("dune-curvedsurfacegrid cannot handle binary-mode .vtu-files. Convert the "
                                 "file to appended-mode or ascii-mode and try again.");
  }else if(connectivity_buf.find("\"appended\"") != npos){
    has_appended_data = true;
    pos1 = connectivity_buf.find("offset=\"");
    if(pos1 == npos) error("Could not find offset for appended data of \"connectivity\".");
    else pos1 += 8;
    pos2 = connectivity_buf.find("\"", pos1);
    if(pos2 == npos) error("Could not find offset for appended data of \"connectivity\".");
    offset_cell_connectivity = ival64(connectivity_buf.substr(pos1, pos2 - pos1));
    pos1 = connectivity_buf.find("type=\"");
    if(pos1 == npos) error("Could not find datatype for appended data of \"connectivity\".");
    else pos1 += 6;
    pos2 = connectivity_buf.find("\"", pos1);
    if(pos2 == npos) error("Could not find datatype for appended data of \"connectivity\".");
    datatype_cell_connectivity = connectivity_buf.substr(pos1, pos2 - pos1);
  }else{
    do{
      int i = 0;
      for(; i <= m.dim_of_elements; ++i){
        in >> connectivity_buf;
        if(connectivity_buf.size() > 0 && connectivity_buf[0] == '<') break;
        v[i] = ival(connectivity_buf);
      }
      if(i <= m.dim_of_elements) break;
      m.add_element(v);
    }while(true);
  }

  //get vertex-vectors
  if(vecs.size() > 0){
    if(position_of_vectors != 0){
      in.seekg(position_of_vectors);
    }else{
      do{
        getline(in, buf);
        if(in.eof()) error("Did not find value-vectors.");
        buf = ltrim(buf);
        if(buf.size() >= 9 && buf.substr(0, 9) == "<CellData"){
          position_of_element_vectors = in.tellg();
        }
      }while(buf.size() < 10 || buf.substr(0,10) != "<PointData");
    }
    for(unsigned e = 0; e < vecs.size(); ++e){
      do{
        getline(in, buf);
        buf = ltrim(buf);
        if(in.eof()){
          if(e < vecs.size()){
            error("Did not find enough value-vectors in file \"" + file
                        + "\"\nNumber of value vectors found: " + ft(e));
          }else{
            return;
          }
        }
      }while(buf.size() < 10 || buf.substr(0, 10) != "<DataArray");
      pos1 = buf.find("NumberOfComponents=\"");
      if(pos1 != npos){
        pos1 += 20;
        pos2 = buf.find("\"", pos1);
        if(pos2 == npos){
          error("\"NumberOfComponents\"-entry for vertex-vector " + ft(e) + " is erroneous.");
        }
        component_count = ival(buf.substr(pos1, pos2 - pos1));
      }else{
        component_count = 1;
      }
      component_counts_point_data.push_back(component_count);
      if(buf.find("\"binary\"") != npos){
        error("dune-curvedsurfacegrid cannot handle binary-mode .vtu-files. Convert the "
                                     "file to appended-mode or ascii-mode and try again.");
      }else if(buf.find("\"appended\"") != npos){
        has_appended_data = true;
        pos1 = buf.find("offset=\"");
        if(pos1 == npos) error("Could not find offset for appended data of \"PointData\".");
        else pos1 += 8;
        pos2 = buf.find("\"", pos1);
        if(pos2 == npos) error("Could not find offset for appended data of \"PointData\".");
        offsets_point_data.push_back(ival64(buf.substr(pos1, pos2 - pos1)));
        pos1 = buf.find("type=\"");
        if(pos1 == npos) error("Could not find datatype for appended data of \"PointData\".");
        else pos1 += 6;
        pos2 = buf.find("\"", pos1);
        if(pos2 == npos) error("Could not find datatype for appended data of \"PointData\".");
        datatypes_point_data.push_back(buf.substr(pos1, pos2 - pos1));
      }else{
        offsets_point_data.push_back(-1);
        unsigned v_count = vertex_count * component_count;
        vecs[e]->resize(v_count);
        for(unsigned i = 0; i < v_count; ++i){
          in >> buf;
          (*vecs[e])[i] = dval(buf);
        }
      }
    }
  }

  //get element-vectors
  if(element_vecs.size() > 0){
    if(position_of_element_vectors != 0){
      in.seekg(position_of_element_vectors);
    }else{
      do{
        getline(in, buf);
        if(in.eof()) error("Did not find element-vectors.");
        buf = ltrim(buf);
      }while(buf.size() < 9 || buf.substr(0, 9) != "<CellData");
    }
    for(unsigned e = 0; e < element_vecs.size(); ++e){
      do{
        getline(in, buf);
        buf = ltrim(buf);
        if(in.eof()){
          if(e < element_vecs.size()){
            error("Did not find enough element-vectors in file \"" + file
                  + "\"\nNumber of element-vectors found: " + ft(e));
          }else{
            return;
          }
        }
      }while(buf.size() < 10 || buf.substr(0, 10) != "<DataArray");
      pos1 = buf.find("NumberOfComponents=\"");
      if(pos1 != npos){
        pos1 += 20;
        pos2 = buf.find("\"", pos1);
        if(pos2 == npos){
          error("\"NumberOfComponents\"-entry for element-vector " + ft(e) + " is erroneous.");
        }
        component_count = ival(buf.substr(pos1, pos2 - pos1));
      }else{
        component_count = 1;
      }
      component_counts_cell_data.push_back(component_count);
      if(buf.find("\"binary\"") != npos){
        error("dune-curvedsurfacegrid cannot handle binary-mode .vtu-files. Convert the "
                                     "file to appended-mode or ascii-mode and try again.");
      }else if(buf.find("\"appended\"") != npos){
        has_appended_data = true;
        pos1 = buf.find("offset=\"");
        if(pos1 == npos) error("Could not find offset for appended data of \"CellData\".");
        else pos1 += 8;
        pos2 = buf.find("\"", pos1);
        if(pos2 == npos) error("Could not find offset for appended data of \"CellData\".");
        offsets_cell_data.push_back(ival64(buf.substr(pos1, pos2 - pos1)));
        pos1 = buf.find("type=\"");
        if(pos1 == npos) error("Could not find datatype for appended data of \"CellData\".");
        else pos1 += 6;
        pos2 = buf.find("\"", pos1);
        if(pos2 == npos) error("Could not find datatype for appended data of \"CellData\".");
        datatypes_cell_data.push_back(buf.substr(pos1, pos2 - pos1));
      }else{
        offsets_cell_data.push_back(-1);
        unsigned v_count = m.elements.size() * component_count;
        element_vecs[e]->resize(v_count);
        for(unsigned i = 0; i < v_count; ++i){
          in >> buf;
          (*element_vecs[e])[i] = dval(buf);
        }
      }
    }
  }

  //read appended data
  if(has_appended_data){
    do{
      getline(in, buf);
      if(in.eof()) error("Could not find appended data.");
      buf = ltrim(buf);
      if(buf.size() >= 32 && buf.substr(0, 32) == "<AppendedData encoding=\"base64\">"){
        error("Base64-encoding in appended-mode vtu-files not supported.");
      }
    }while(buf.size() < 29 || buf.substr(0, 29) != "<AppendedData encoding=\"raw\">");
    char tmp = ' ';
    do{
      in.read(&tmp, 1);
    }while(tmp != '_');
    int64_t offset_appended_data = in.tellg();
    VtuValueReader *vvr = has_compression? new VtuValueDecompressor(in, header_size)
                                         : new VtuValueReader(in, header_size);

    //appended vertices
    if(offset_points != -1){
      in.seekg(offset_appended_data + offset_points, std::ios_base::beg);
      vvr->initialize();
      if(datatype_points == "Float32"){
        unsigned point_count = vvr->total_size / 12;
        if(sizeof(float) != 4) error("Size of type float is not 4 bytes!");
        for(unsigned i = 0; i < point_count; ++i){
          vc[0] = vvr->read_float_value();
          vc[1] = vvr->read_float_value();
          vc[2] = vvr->read_float_value();
          m.add_vertex(vc);
        }
      }else if(datatype_points == "Float64"){
        unsigned point_count = vvr->total_size / 24;
        if(sizeof(double) != 8) error("Size of type double is not 8 bytes!");
        for(unsigned i = 0; i < point_count; ++i){
          vc[0] = vvr->read_double_value();
          vc[1] = vvr->read_double_value();
          vc[2] = vvr->read_double_value();
          m.add_vertex(vc);
        }
      }else{
        error("Bad datatype \"" + datatype_points + "\" for Points.");
      }
    }

    //appended cell-types
    if(offset_cell_types != -1){
      in.seekg(offset_appended_data + offset_cell_types, std::ios_base::beg);
      vvr->initialize();
      if(datatype_cell_types == "UInt8"){
        if(vvr->total_size == 0){
          m.dim_of_elements = 2;
        }else{
          uchar uc = vvr->read_uchar_value();
          if(uc == 3) m.dim_of_elements = 1;
          else if(uc == 5) m.dim_of_elements = 2;
          else if(uc == 10) m.dim_of_elements = 3;
          else error("Unknown element type " + ft(int(uc)) + ".");
          for(unsigned e = 1, type_count = vvr->total_size; e < type_count; ++e){
            uchar tmp_uc = vvr->read_uchar_value();
            if(tmp_uc != uc){
              error("File contains multiple element types! At least "
                      + ft(int(uc)) + " and " + ft(int(tmp_uc)) + ".");
            }
          }
        }
      }else{
        error("Bad datatype \"" + datatype_cell_types + "\" for cell types.");
      }
    }

    //appended cell-connectivity
    if(offset_cell_connectivity != -1){
      in.seekg(offset_appended_data + offset_cell_connectivity, std::ios_base::beg);
      vvr->initialize();
      if(datatype_cell_connectivity == "Int64"){
        unsigned element_count = vvr->total_size / ((m.dim_of_elements + 1) * 8);
        for(unsigned e = 0; e < element_count; ++e){
          for(int i = 0; i <= m.dim_of_elements; ++i){
            v[i] = vvr->read_int64_value();
          }
          m.add_element(v);
        }
      }else if(datatype_cell_connectivity == "Int32"){
        unsigned element_count = vvr->total_size / ((m.dim_of_elements + 1) * 4);
        for(unsigned e = 0; e < element_count; ++e){
          for(int i = 0; i <= m.dim_of_elements; ++i){
            v[i] = vvr->read_int32_value();
          }
          m.add_element(v);
        }
      }else{
        error("Bad datatype \"" + datatype_cell_connectivity + "\" for cell connectivity.");
      }
    }

    //appended vertex-vectors
    for(unsigned e = 0; e < offsets_point_data.size(); ++e){
      if(offsets_point_data[e] != -1){
        in.seekg(offset_appended_data + offsets_point_data[e], std::ios_base::beg);
        vvr->initialize();
        if(datatypes_point_data[e] == "Float32"){
          unsigned entry_count = vvr->total_size / 4;
          if(int(entry_count / component_counts_point_data[e]) != m.vertices.size()){
            error("Number of vertices does not match length of appended data for vertex-vector "
                                                                                   + ft(e) + ".");
          }
          if(sizeof(float) != 4) error("Size of type float is not 4 bytes!");
          vecs[e]->resize(entry_count);
          for(unsigned i = 0; i < entry_count; ++i){
            (*vecs[e])[i] = vvr->read_float_value();
          }
        }else if(datatypes_point_data[e] == "Float64"){
          unsigned entry_count = vvr->total_size / 8;
          if(int(entry_count / component_counts_point_data[e]) != m.vertices.size()){
            error("Number of vertices does not match length of appended data for vertex-vector "
                                                                                   + ft(e) + ".");
          }
          if(sizeof(double) != 8) error("Size of type double is not 8 bytes!");
          vecs[e]->resize(entry_count);
          for(unsigned i = 0; i < entry_count; ++i){
            (*vecs[e])[i] = vvr->read_double_value();
          }
        }else if(datatypes_point_data[e] == "Int64"){
          unsigned entry_count = vvr->total_size / 8;
          if(int(entry_count / component_counts_point_data[e]) != m.vertices.size()){
            error("Number of vertices does not match length of appended data for vertex-vector "
                                                                                   + ft(e) + ".");
          }
          vecs[e]->resize(entry_count);
          for(unsigned i = 0; i < entry_count; ++i){
            (*vecs[e])[i] = vvr->read_int64_value();
          }
        }else{
          error("Bad datatype \"" + datatypes_point_data[e] + "\" for point data.");
        }
      }
    }

    //appended element-vectors
    for(unsigned e = 0; e < offsets_cell_data.size(); ++e){
      if(offsets_cell_data[e] != -1){
        in.seekg(offset_appended_data + offsets_cell_data[e], std::ios_base::beg);
        vvr->initialize();
        if(datatypes_cell_data[e] == "Float32"){
          unsigned entry_count = vvr->total_size / 4;
          if(int(entry_count / component_counts_cell_data[e]) != m.elements.size()){
            error("Number of elements does not match length of appended data for element-vector "
                                                                                    + ft(e) + ".");
          }
          if(sizeof(float) != 4) error("Size of type float is not 4 bytes!");
          element_vecs[e]->resize(entry_count);
          for(unsigned i = 0; i < entry_count; ++i){
            (*element_vecs[e])[i] = vvr->read_float_value();
          }
        }else if(datatypes_cell_data[e] == "Float64"){
          unsigned entry_count = vvr->total_size / 8;
          if(int(entry_count / component_counts_cell_data[e]) != m.elements.size()){
            error("Number of elements does not match length of appended data for element-vector "
                                                                                    + ft(e) + ".");
          }
          if(sizeof(double) != 8) error("Size of type double is not 8 bytes!");
          element_vecs[e]->resize(entry_count);
          for(unsigned i = 0; i < entry_count; ++i){
            (*element_vecs[e])[i] = vvr->read_double_value();
          }
        }else if(datatypes_cell_data[e] == "Int64"){
          unsigned entry_count = vvr->total_size / 8;
          if(int(entry_count / component_counts_cell_data[e]) != m.elements.size()){
            error("Number of elements does not match length of appended data for element-vector "
                                                                                    + ft(e) + ".");
          }
          element_vecs[e]->resize(entry_count);
          for(unsigned i = 0; i < entry_count; ++i){
            (*element_vecs[e])[i] = vvr->read_int64_value();
          }
        }else{
          error("Bad datatype \"" + datatypes_cell_data[e] + "\" for cell data.");
        }
      }
    }
    delete vvr;
  }

  in.close();
}
