// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef COORDFUNCTION

#include <dune/common/timer.hh>

#include <dune/curvedsurfacegrid/capabilities.hh>
#include <dune/curvedsurfacegrid/curvedsurfacegrid.hh>
#include <dune/curvedsurfacegrid/coordfunctionimplementations.hh>
// #include <dune/grid/io/file/dgfparser.hh>
#if HAVE_UG
#include <dune/grid/albertagrid.hh>
// #include <dune/grid/albertagrid/dgfparser.hh>
#endif
#include <dune/grid/yaspgrid.hh>
// #include <dune/grid/io/file/dgfparser/dgfyasp.hh>
// #include <dune/grid/io/file/dgfparser/dgfgeogrid.hh>
#include <dune/grid/utility/structuredgridfactory.hh>
#include <dune/grid/test/functions.hh>

#include <dune/grid/test/gridcheck.hh>
#include <dune/grid/test/checkcommunicate.hh>
#include <dune/grid/test/checkgeometryinfather.hh>
#include <dune/grid/test/checkintersectionit.hh>
#include <dune/grid/test/checkiterators.hh>
#include <dune/grid/test/checkpartition.hh>
#include <dune/grid/test/checkgeometry.hh>

#include <dune/functions/common/differentiablefunctionfromcallables.hh>
#include <dune/functions/common/signature.hh>
#include <dune/functions/gridfunctions/analyticgridviewfunction.hh>

namespace Dune
{
  template< int dim, int dimworld >
  class AlbertaGrid;
}

using namespace Dune;

template< int dim, int dimworld >
struct EnableLevelIntersectionIteratorCheck< AlbertaGrid< dim, dimworld > >
{
  static const bool v = false;
};

template< class GridFunction, int order >
struct EnableLevelIntersectionIteratorCheck< CurvedSurfaceGrid< GridFunction, order > >
{
  using HostGrid = typename GridFunction::GridView::Grid;
  static const bool v = EnableLevelIntersectionIteratorCheck< HostGrid >::v;
};

typedef GRIDTYPE HostGrid;
static const int dow = HostGrid::dimensionworld;

// define an identity coordinate mapping
struct Identity
{
  FieldVector<double,dow> operator() (const FieldVector<double,dow>& x) const
  {
    return x;
  }
};

// define the derivative of an identity coorinate mapping
struct DIdentity
{
  FieldMatrix<double,dow,dow> operator() (const FieldVector<double,dow>& x) const
  {
    FieldMatrix<double,dow,dow> I(0);
    for (int i = 0; i < dow; ++i)
      I[i][i] = 1;
    return I;
  }
};

// construct a differentiable callable from two callables
using IdentityFunction = Functions::DifferentiableFunctionFromCallables<
  FieldVector<double,dow>(FieldVector<double,dow>),
  Functions::DefaultDerivativeTraits,
  Identity,
  DIdentity>;

// Construct a GridFunction from the differentiable callable
using IdentityGridFunction = Functions::AnalyticGridViewFunction<
  FieldVector<double,dow>(FieldVector<double,dow>),
  typename HostGrid::LeafGridView,
  IdentityFunction>;

using GridFunction = IdentityGridFunction;


template <class GridType>
void test()
{
  using SingleGeometry = Dune::Capabilities::hasSingleGeometryType<GridType>;
  bool useSimplices = Dune::GeometryType(SingleGeometry::topologyId, GridType::dimension, false).isSimplex();
  auto gridPtr = useSimplices
    ? Dune::StructuredGridFactory<HostGrid>::createSimplexGrid({0.0, 0.0}, {1.0, 1.0}, {1u, 1u})
    : Dune::StructuredGridFactory<HostGrid>::createCubeGrid({0.0, 0.0}, {1.0, 1.0}, {1u, 1u});

  IdentityFunction idFct(Identity{}, DIdentity{});
  IdentityGridFunction idGridFct(idFct, gridPtr->leafGridView());
  GridType geogrid(*gridPtr, idGridFct);

  geogrid.globalRefine( 1 );
  geogrid.loadBalance();

  std::cerr << "Checking grid..." << std::endl;
  gridcheck( geogrid );

  std::cerr << "Checking geometry... " << std::endl;
  Dune::GeometryChecker< GridType > checker;

  checker.checkGeometry( geogrid.leafGridView() );
  for( int i = 0; i <= geogrid.maxLevel(); ++i )
    checker.checkGeometry( geogrid.levelGridView( i ) );

  std::cerr << "Checking geometry in father..." << std::endl;
  checkGeometryInFather( geogrid );
  std::cerr << "Checking intersections..." << std::endl;
  checkIntersectionIterator( geogrid, !EnableLevelIntersectionIteratorCheck< HostGrid >::v );

  checkIterators( geogrid.leafGridView() );
  for( int i = 0; i <= geogrid.maxLevel(); ++i )
    checkIterators( geogrid.levelGridView( i ) );

  checkPartitionType( geogrid.leafGridView() );
  for( int i = 0; i <= geogrid.maxLevel(); ++i )
    checkPartitionType( geogrid.levelGridView( i ) );

  std::cerr << "Checking geometry lifetime..." << std::endl;
  checkGeometryLifetime( geogrid.leafGridView() );

  std::cerr << "Checking communication..." << std::endl;
  checkCommunication( geogrid, -1, std::cout );
  if( EnableLevelIntersectionIteratorCheck< HostGrid >::v )
  {
    for( int i = 0; i <= geogrid.maxLevel(); ++i )
      checkCommunication( geogrid, i, std::cout );
  }

}

#if 0
void testNestedCurvedSurfaceGrid(const std::string& gridfile) {

  using NestedCurvedSurfaceGrid
    = Dune::CurvedSurfaceGrid< CurvedSurfaceGrid, Dune::IdenticalCoordFunction< Grid::ctype, dow >, 2 >;
  using NestedGgLeafView = NestedCurvedSurfaceGrid::LeafGridView;

  using SingleGeometry = Dune::Capabilities::hasSingleGeometryType<NestedCurvedSurfaceGrid>;
  bool useSimplices = Dune::GeometryType(SingleGeometry::topologyId, NestedCurvedSurfaceGrid::dimension, false).isSimplex();
  auto gridPtr = useSimplices
    ? Dune::StructuredGridFactory<NestedCurvedSurfaceGrid>::createSimplexGrid({0.0, 0.0}, {1.0, 1.0}, {4u, 4u})
    : Dune::StructuredGridFactory<NestedCurvedSurfaceGrid>::createCubeGrid({0.0, 0.0}, {1.0, 1.0}, {4u, 4u});

  NestedCurvedSurfaceGrid &geogrid = *gridPtr;
  geogrid.globalRefine( 1 );

  // creating different variables for storing grid views
  NestedGgLeafView gv1 = geogrid.leafGridView();
  std::shared_ptr<NestedGgLeafView> pgv2;
  NestedGgLeafView gv3 = geogrid.leafGridView();

  // use geometry grid views copy constructor and copy assignment operator
  {
    NestedGgLeafView tmpGv = geogrid.leafGridView();
    // copying/assigning temporary grid view would be dangerous with default implementation of copy constructor or copy assignment operator from geomrtry grid view
    pgv2 = std::make_shared<NestedGgLeafView> (tmpGv);
    gv3 = tmpGv;
  }
  // The following test would now be guaranteed to fail with the default implementation of geometry grid views copy constructor
  // assert(&(pgv2->indexSet().hostIndexSet()) == &(pgv2->impl().hostGridView().indexSet()))
  // but this would not compile since the member functions impl() and hostIndexSet() are protected/private.
  // So we do the following

  // get index sets and do some arbitrary test to check whether they work correctly
  auto const& is1 = gv1.indexSet();
  auto const& is2 = pgv2->indexSet();
  auto const& is3 = gv3.indexSet();
  for(auto const& e : elements(gv1)) {
    bool idxCorrect = (is1.index(e)==is2.index(e));
    idxCorrect &= (is1.index(e)==is3.index(e));
    assert(idxCorrect);
  }
}
#endif

int main ( int argc, char **argv )
try
{
  Dune::MPIHelper::instance( argc, argv );

  // std::string gridfile = DUNE_GRID_EXAMPLE_GRIDS_PATH "dgf/cube-2.dgf";
  // if (argc >= 2)
  //   gridfile = argv[1];

  Dune::Timer watch;

  watch.reset();
  test<CurvedSurfaceGrid<GridFunction>>();
  std::cout << "=== CurvedSurfaceGrid took " << watch.elapsed() << " seconds\n";

  // watch.reset();
  // testNestedCurvedSurfaceGrid(gridfile);
  // std::cout << "=== NestedCurvedSurfaceGrid took " << watch.elapsed() << " seconds\n";

  // // Check with a discrete coordinate function
  // Dune::GridPtr<Grid> hostGridPtr(gridfile);
  // auto& hostGrid = *hostGridPtr;
  // const int dim = Grid::dimension;

  // typedef Grid::LeafGridView HostGridView;
  // const HostGridView& hostGridView = hostGrid.leafGridView();

  // std::vector<Dune::FieldVector<double,dim> > vertexPositions(hostGridView.size(dim));

  // for (const auto& vertex : vertices(hostGridView))
  //   vertexPositions[hostGridView.indexSet().index(vertex)] = vertex.geometry().corner(0);

  // typedef Dune::GeometryGrid<Grid, DeformationFunction<HostGridView> > DiscretelyTransformedGrid;
  // DeformationFunction<HostGridView> discreteTransformation(hostGridView, vertexPositions);
  // DiscretelyTransformedGrid discretelyTransformedGrid(hostGrid, discreteTransformation);

  // gridcheck(discretelyTransformedGrid);

  return 0;
}
catch( const Dune::Exception &e )
{
  std::cerr << e << std::endl;
  return 1;
}
catch( const std::exception &e )
{
  std::cerr << e.what() << std::endl;
  return 1;
}
catch( ... )
{
  std::cerr << "Unknown exception raised." << std::endl;
  return 1;
}

#else
#error "COORDFUNCTION not defined (e.g., Helix, Circle; see functions.hh)"
#endif
